class LabelsCommon:
    date = "date"
    isoweek = "isoweek"
    month = "month"
    day = "day"


class LabelsExog:
    lbl_calendar_numeric = "calendar_numeric"
    lbl_spline_dim = "spline_dim"
    lbl_is_holiday = "is_holiday"


class LabelsHolidaysOut:
    lbl_country = "country"
    lbl_holiday = "holiday"
